/***********************************************************************/
/* This program is meant to run as an object.                          */
/* It is a simple program to demonstrate:                              */
/*   -- "Private" attributes available to it                           */
/*   -- "Public" methods available to other programs.                  */
/*   -- A good template for future objects with Constructor and        */
/*      Destructor routines.                                           */
/***********************************************************************/

FUNCTION cGetAttr RETURNS CHARACTER (INPUT cAttrName AS CHARACTER) FORWARD.


/* ----------------------- Begin Attributes List ----------------------*/

/* Example attribute specific to the object */

/* Used for object management */
DEFINE VARIABLE cgObjName          as character NO-UNDO.

DEFINE VARIABLE iIterOrder AS INTEGER NO-UNDO.

DEFINE TEMP-TABLE ttAttributes
  FIELD AttrName AS CHARACTER
  FIELD AttrValue AS CHARACTER
  INDEX key1 AttrName ASCENDING.

DEFINE TEMP-TABLE List 
  FIELD Order AS INTEGER INITIAL ?
  FIELD PrevOrder AS INTEGER INITIAL ?
  FIELD NextOrder AS INTEGER INITIAL ?
  FIELD Data AS CHARACTER INITIAL ?
  INDEX pukey Order ASCENDING
  INDEX key1 PrevOrder ASCENDING
  INDEX key2 NextOrder ASCENDING.

/* ------------------------- End Attributes List ----------------------*/

/* -------------------------- Begin Methods List ----------------------*/

/***********************************************************************/
/* This is the "destructor" for the routine.  It should be called be   */
/* fore deleting the handle to the instance of this object.            */
/***********************************************************************/

PROCEDURE Destroy:

END.

/***********************************************************************/
/* If other "constructors" are needed, they can be put in.  This one   */
/* is called Init.  Unlike C++, it will need to be run manually.       */
/* If you are using ObjMgr.i, it will be run automatically.            */
/* If you need to make other Inits, place them here and name beginning */
/* "Init" : InitByRowID or InitBySalesOrderNumber.                     */
/***********************************************************************/

PROCEDURE Init:

  DEFINE INPUT PARAMETER cName AS CHARACTER NO-UNDO.

  ASSIGN cgObjName = cName.

  RUN SetError IN THIS-PROCEDURE ("000", "Init").

  ASSIGN iIterOrder = 0.

  RUN ClearList IN THIS-PROCEDURE.

END.

/***********************************************************************/
/* All procedures need a way to describe their errors back to the cal- */
/* ler.                                                                */
/***********************************************************************/

PROCEDURE GetError:

  DEFINE OUTPUT PARAMETER cErrCode AS CHARACTER NO-UNDO.
  DEFINE OUTPUT PARAMETER cErrMsg  AS CHARACTER NO-UNDO.
  DEFINE OUTPUT PARAMETER cErrMethod AS CHARACTER NO-UNDO.

  ASSIGN
  cErrCode = cGetAttr("ErrCode")
  cErrMsg  = cGetAttr("ErrMsg")
  cErrMethod = cGetAttr("ErrMethod").

END. /* PROCEDURE GetError */

/***********************************************************************/
/* This is the central point where internal procedures can communicate */
/* their problems to the procedure as a whole.                         */
/* It's main purpose is to convert codes into human readable form in   */
/* cgObjErrMsg as well as populate the procedures globally available   */
/* vars.                                                               */
/***********************************************************************/

PROCEDURE SetError:

  DEFINE INPUT PARAMETER cErrorCode AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER cMethodName AS CHARACTER NO-UNDO.

  CASE cErrorCode:

    WHEN "000" THEN DO:
      RUN SetAttr IN THIS-PROCEDURE ("ErrCode", cErrorCode).
      RUN SetAttr IN THIS-PROCEDURE ("ErrMsg", "No Error").
      RUN SetAttr IN THIS-PROCEDURE ("ErrMethod", cMethodName).
    END.

    OTHERWISE DO:
      RUN SetAttr IN THIS-PROCEDURE ("ErrCode", cErrorCode).
      RUN SetAttr IN THIS-PROCEDURE ("ErrMsg", ?).
      RUN SetAttr IN THIS-PROCEDURE ("ErrMethod", cMethodName).
    END.

  END. /* CASE */

END. /* PROCEDURE SetError */

/***********************************************************************/
/* These are methods to perform activities on the data controlled by   */
/* the object.                                                         */
/* Attributes are actually stored in a temp-table, that way we can add */
/* new ones easily, and not need to create more Set/Get routines for   */
/* for each attribute.  The bad news is, we need to cast to and from   */
/* character for the data to pass back and fourth.  If this is a pron- */
/* lem, then create some Set/Get routines for those values.            */
/***********************************************************************/


/***********************************************************************/
/* We don't make this a function, because we need to FORWARD declare   */
/* in the code using this object, and this name is going to be pret-   */
/* ty popular causing conflict with other objects with a GetAttr.      */
/***********************************************************************/

PROCEDURE GetAttr:

  DEFINE INPUT PARAMETER cName   AS CHARACTER NO-UNDO.
  DEFINE OUTPUT PARAMETER cValue AS CHARACTER NO-UNDO.

  FIND ttAttributes NO-LOCK
  WHERE ttAttributes.AttrName = cName
  NO-ERROR.

  IF NOT AVAILABLE ttAttributes THEN DO:

    /*********************************************************/
    /* Note that sometimes it is OK for an attribute to be ? */
    /* so be sure to remember to check the error if the at-  */
    /* tribute wasn't found or really is ?.  Sometimes it    */
    /* it works out it does not matter, sometimes it does    */
    /* matter.                                               */
    /*********************************************************/

    RUN SetError IN THIS-PROCEDURE ("001", "GetAttr").
    ASSIGN cValue = ?.
    RETURN.

  END. /* IF NOT AVAILABLE ttAttributes */

  ASSIGN cValue = ttAttributes.AttrValue.

END.  /* PROCEDURE GetAttr */

/***********************************************************************/
/* However it IS useful to have a GetAttr function for use in THIS-    */
/* PROCEDURE within the internal procedures available.                 */
/***********************************************************************/

FUNCTION cGetAttr RETURNS CHARACTER (INPUT cAttrName AS CHARACTER):

  DEFINE VARIABLE cAttrValue AS CHARACTER NO-UNDO.

  RUN GetAttr IN THIS-PROCEDURE (INPUT cAttrName, OUTPUT cAttrValue).

  RETURN cAttrValue.

END. /* FUNCTION cGetAttr() */

/***********************************************************************/
/* If the attribute already exists, we overwrite, not error out...     */
/***********************************************************************/

PROCEDURE SetAttr:

  DEFINE INPUT PARAMETER cName   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER cValue AS CHARACTER NO-UNDO.

  FIND ttAttributes NO-LOCK
  WHERE ttAttributes.AttrName = cName
  NO-ERROR.

  IF NOT AVAILABLE ttAttributes THEN CREATE ttAttributes.

  ASSIGN ttAttributes.AttrName = cName
         ttAttributes.AttrValue = cValue.

  /* Some attributes are dependent on other attributes, handle those */
  /* here.                                                           */

  RUN DependentAttr (cName, cValue).

END. /* PROCEDURE SetAttr */

/***********************************************************************/
/* Some attributes are dependent on the values of other attributes.    */
/* We keep them in sync with this code here.                           */
/***********************************************************************/

PROCEDURE DependentAttr:

  DEFINE INPUT PARAMETER cName   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER cValue AS CHARACTER NO-UNDO.

END. /* PROCEDURE DependentAttr */

/***********************************************************************/
/* This is a way to quickly transfer record information into attri-    */
/* butes.  The attribute name is tablename_fieldname.                  */
/* Example Use:                                                        */
/* FIND FIRST Person NO-LOCK.                                          */
/* ASSIGN hBuffer = BUFFER Person:Handle.                              */
/* RUN Record2Attr (hBuffer).                                          */
/***********************************************************************/

PROCEDURE Record2Attr:

  DEFINE INPUT PARAMETER hBuffer AS HANDLE NO-UNDO.

  DEFINE VARIABLE hField     AS HANDLE NO-UNDO.
  DEFINE VARIABLE i          AS INTEGER NO-UNDO.

  DO i = 1 TO hBuffer:Num-Fields:

    ASSIGN hField = hBuffer:Buffer-Field(i).

    RUN SetAttr IN THIS-PROCEDURE (hBuffer:Name + "_" + hField:Name, hField:String-Value).

  END.

END.

/***********************************************************************/
/* Scan the attributes table for entries beginning with the table name */
/* and apply their values to the field named in the second part of the */
/* attribute name.                                                     */
/* Note this doesn't manage multiple buffers of the same name very     */
/* well.  If you need n records from the same table, name the buffers  */
/* seperately.                                                         */
/***********************************************************************/

PROCEDURE Attr2Record:

  DEFINE INPUT PARAMETER hBuffer AS HANDLE NO-UNDO.

  DEFINE VARIABLE hField     AS HANDLE NO-UNDO.
  DEFINE VARIABLE cFieldName AS CHARACTER NO-UNDO.

  FOR EACH ttAttributes NO-LOCK
  WHERE ttAttributes.AttrName BEGINS hBuffer:Name + "_":

    /*****************************************************************/
    /* Luckily, it appears that buffer-value types automatically...  */
    /* It will puke on bad data sent - such as text for an int, etc. */
    /*****************************************************************/

    ASSIGN hfield = hBuffer:Buffer-Field(ENTRY(2, ttAttributes.AttrName, "_"))
           hfield:Buffer-Value = ttAttributes.AttrValue.

  END. /* FOR EACH ttAttributes */

END. /* PROCEDURE Attr2Record */


/***********************************************************************/
/* Useful for debugging. Note we output to a file so we don't get any  */
/* wrong display device type errors when the r-code is shared between  */
/* different interfaces.                                               */
/***********************************************************************/

PROCEDURE AttrDebug:

  DEFINE INPUT PARAMETER cFileName AS CHARACTER NO-UNDO.

  OUTPUT TO VALUE(cFileName).

  FOR EACH ttAttributes NO-LOCK:

    PUT UNFORMATTED ttAttributes.AttrName "=" ttAttributes.AttrValue SKIP.

  END. /* FOR EACH ttAttributes */

  OUTPUT CLOSE.

END. /* PROCEDURE AttrDebug */

/***********************************************************************/
/* Clear out all attributes that have been set.  It is better to set   */
/* an attribute to ? and code for that; but sometimes one needs this.  */
/***********************************************************************/

PROCEDURE ClearAttr:

  FOR EACH ttAttributes:

    DELETE ttAttributes.

  END. /* FOR EACH ttAttributes */

END.  /* PROCEDURE ClearAttr */

/***********************************************************************/
/* This is a special set of routines that allow the object's user to   */
/* reach the tables directly.  Convenient for ASSIGNS and the like so  */
/* one is not continously calling a Set* routine for each and every    */
/* field.                                                              */
/* WARNING: This CAN BE ABUSED.                                        */
/* WARNING: Sets of Like records may be a problem.                     */
/*    ATTN: Buffers can go both ways.  This can be good.  This can be  */
/*          bad.                                                       */
/***********************************************************************/
/*
PROCEDURE GetBuffers:

  DEFINE PARAMETER BUFFER YourRecord FOR YourTable.

  BUFFER-COPY ObjBuffer YourRecord.

END. /* PROCEDURE GetBuffers */

PROCEDURE SetBuffers:

  DEFINE PARAMETER BUFFER YourRecord FOR YourTable.

  BUFFER-COPY YourRecord ObjBuffer.

END.
*/


DEFINE VARIABLE iCurOrder  AS INTEGER NO-UNDO.

PROCEDURE GoFront:

  FIND List NO-LOCK 
  WHERE List.PrevOrder = ?
  NO-ERROR.

  IF AVAILABLE List THEN ASSIGN iCurOrder = List.Order.

END. /* PROCEDURE GoFront */

PROCEDURE GoRear:

  FIND List NO-LOCK 
  WHERE List.NextOrder = ?
  NO-ERROR.

  IF AVAILABLE List THEN ASSIGN iCurOrder = List.Order.

END. /* PROCEDURE GoRear */

PROCEDURE GoNext:

  FIND List NO-LOCK
  WHERE List.Order = iCurOrder
  NO-ERROR.

  IF NOT AVAILABLE List THEN RETURN.

  IF List.NextOrder = ? THEN RETURN.

  ASSIGN iCurOrder = List.NextOrder.
  
END. /* PROCEDURE GoNext */

PROCEDURE GoPrev:

  FIND List NO-LOCK
  WHERE List.Order = iCurOrder
  NO-ERROR.

  IF NOT AVAILABLE List THEN RETURN.

  IF List.PrevOrder = ? THEN RETURN.

  ASSIGN iCurOrder = List.PrevOrder.

END. /* PROCEDURE GoPrev */

PROCEDURE JumpTo:

  DEFINE INPUT PARAMETER iJumpLoc AS INTEGER NO-UNDO.

  ASSIGN iCurOrder = iJumpLoc.

END. /* PROCEDURE JumpTo */

PROCEDURE GetLoc:

  DEFINE OUTPUT PARAMETER iLoc AS INTEGER NO-UNDO.

  ASSIGN iLoc = iCurOrder.

END. /* PROCEDURE GetLoc */

PROCEDURE FindLoc:
  
  DEFINE INPUT PARAMETER cFindString AS CHARACTER NO-UNDO.
  DEFINE OUTPUT PARAMETER iLoc AS INTEGER NO-UNDO.

  FOR EACH List NO-LOCK:
    IF List.Data = cFindString THEN DO:
      ASSIGN iLoc = List.Order.
      RETURN.
    END.
  END.

  ASSIGN iLoc = ?.

END. /* PROCEDURE FindLoc */

PROCEDURE Remove:

  DEFINE BUFFER Prev_List FOR List.
  DEFINE BUFFER Next_List FOR List.
  DEFINE BUFFER Cur_List FOR List.

  FIND Cur_List NO-LOCK
  WHERE Cur_List.Order = iCurOrder
  NO-ERROR.

  IF NOT AVAILABLE Cur_List THEN RETURN.

  IF Cur_List.NextOrder <> ? THEN
    FIND Next_List EXCLUSIVE-LOCK
    WHERE Next_List.Order = Cur_List.NextOrder
    NO-ERROR.

  IF Cur_List.PrevOrder <> ? THEN
    FIND Prev_List EXCLUSIVE-LOCK
    WHERE Prev_List.Order = Cur_List.PrevOrder
    NO-ERROR.

  DELETE Cur_List.

  IF AVAILABLE Prev_List AND AVAILABLE Next_List THEN DO:

    ASSIGN Prev_List.NextOrder = Next_List.Order
           Next_List.PrevOrder = Prev_List.Order
           iCurOrder = Next_List.Order.

  END. /* AVAILABLE Prev_List AND AVAILABLE Next_List */

  ELSE IF AVAILABLE Prev_List AND NOT AVAILABLE Next_List THEN DO: 

    ASSIGN Prev_List.NextOrder = ?
           iCurOrder = Prev_List.Order.

  END.

  ELSE IF NOT AVAILABLE Prev_List AND AVAILABLE Next_List THEN DO:

    ASSIGN Next_List.PrevOrder = ?
           iCurOrder = Next_List.Order.

  END.

END. /* PROCEDURE Remove */

PROCEDURE AddBefore:

  DEFINE INPUT PARAMETER cValue AS CHARACTER NO-UNDO.

  DEFINE BUFFER Prev_List FOR List.
  DEFINE BUFFER Cur_List FOR List.
  DEFINE BUFFER New_List FOR List.

  FIND Cur_List NO-LOCK
  WHERE Cur_List.Order = iCurOrder
  NO-ERROR.

  IF NOT AVAILABLE Cur_List THEN DO:

    CREATE New_List.
    ASSIGN iIterOrder = iIterOrder + 1
           New_List.Order = iIterOrder
           iCurOrder = iIterOrder
           New_List.Data = cValue
           New_List.NextOrder = ?
           New_List.PrevOrder = ?.

    RETURN.

  END. /* NOT AVAILABLE Cur_List */

  FIND Prev_List NO-LOCK
  WHERE Prev_List.Order = Cur_List.PrevOrder
  NO-ERROR.

  CREATE New_List.

  ASSIGN iIterOrder = iIterOrder + 1
         New_List.Order = iIterOrder
         iCurOrder = iIterOrder.

  IF AVAILABLE Prev_List THEN DO:
    
    ASSIGN Prev_List.NextOrder = New_List.Order
           New_List.PrevOrder = Prev_List.Order
           New_List.NextOrder = Cur_List.Order
           Cur_List.PrevOrder = New_List.Order.
  END.

  ELSE IF NOT AVAILABLE Prev_List THEN DO:

    ASSIGN New_List.PrevOrder = ?
           New_List.NextOrder = Cur_List.Order
           Cur_List.PrevOrder = New_List.Order.
  END.

  ASSIGN New_List.Data = cValue.

END. /* PROCEDURE AddBefore */

PROCEDURE AddAfter:

  DEFINE INPUT PARAMETER cValue AS CHARACTER NO-UNDO.

  DEFINE BUFFER Next_List FOR List.
  DEFINE BUFFER Cur_List FOR List.
  DEFINE BUFFER New_List FOR List.

  FIND Cur_List NO-LOCK
  WHERE Cur_List.Order = iCurOrder
  NO-ERROR.

  IF NOT AVAILABLE Cur_List THEN DO:

    CREATE New_List.
    ASSIGN iIterOrder = iIterOrder + 1
           New_List.Order = iIterOrder
           iCurOrder = iIterOrder
           New_List.Data = cValue
           New_List.NextOrder = ?
           New_List.PrevOrder = ?.

    RETURN.

  END. /* NOT AVAILABLE Cur_List */

  FIND Next_List NO-LOCK
  WHERE Next_List.Order = Cur_List.NextOrder
  NO-ERROR.

  CREATE New_List.

  ASSIGN iIterOrder = iIterOrder + 1
         New_List.Order = iIterOrder.

  IF AVAILABLE Next_List THEN DO:

    ASSIGN Next_List.PrevOrder = New_List.Order
           New_List.PrevOrder = Cur_List.Order
           New_List.NextOrder = Next_List.Order
           Cur_List.NextOrder = New_List.Order.
  END.

  ELSE IF NOT AVAILABLE Next_List THEN DO:

    ASSIGN New_List.NextOrder = ?
           New_List.PrevOrder = Cur_List.Order
           Cur_List.NextOrder = New_List.Order.
  END.

  ASSIGN New_List.Data = cValue.

END. /* PROCEDURE AddAfter */

PROCEDURE AddFront:

  DEFINE INPUT PARAMETER cValue AS CHARACTER NO-UNDO.

  RUN GoFront IN THIS-PROCEDURE.
  RUN AddBefore IN THIS-PROCEDURE (cValue).

END. /* PROCEDURE AddFront */

PROCEDURE AddRear:

  DEFINE INPUT PARAMETER cValue AS CHARACTER NO-UNDO.

  RUN GoRear IN THIS-PROCEDURE.
  RUN AddAfter IN THIS-PROCEDURE (cValue).

END. /* PROCEDURE AddRear */

PROCEDURE ReadLoc:

  DEFINE INPUT PARAMETER iJumpLoc AS INTEGER NO-UNDO.
  DEFINE OUTPUT PARAMETER cValue AS CHARACTER NO-UNDO.

  RUN JumpTo IN THIS-PROCEDURE (iJumpLoc).
  RUN ReadValue IN THIS-PROCEDURE (OUTPUT cValue).

END. /* PROCEDURE ReadLoc */

PROCEDURE ReadValue:

  DEFINE OUTPUT PARAMETER cValue AS CHARACTER NO-UNDO.

  FIND List NO-LOCK
  WHERE List.Order = iCurOrder
  NO-ERROR.

  IF AVAILABLE List THEN ASSIGN cValue = List.Data.
  ELSE ASSIGN cValue = ?.

END. /* PROCEDURE ReadValue */

PROCEDURE Count:

  DEFINE OUTPUT PARAMETER iCount AS INTEGER NO-UNDO.

  ASSIGN iCount = 0.
  FOR EACH List NO-LOCK:
    ASSIGN iCount = iCount + 1.
  END.

END. /* PROCEDURE Count */

PROCEDURE IsEmpty:

  DEFINE OUTPUT PARAMETER lIsEmpty AS LOGICAL NO-UNDO.

  FIND FIRST List NO-LOCK NO-ERROR.

  lIsEmpty = AVAILABLE List.

END. /* PROCEDURE IsEmpty */

PROCEDURE DumpList:

  DEFINE INPUT PARAMETER cFileName AS CHARACTER NO-UNDO.

  OUTPUT TO VALUE(cFileName).

  FOR EACH List:

    EXPORT List.

  END. /* FOR EACH List */

  PUT UNFORMATTED "iCurOrder = " iCurOrder SKIP
                  "iIterOrder = " iIterOrder SKIP.
  OUTPUT CLOSE.

END. /* PROCEDURE DumpList */

PROCEDURE ClearList:

  FOR EACH List EXCLUSIVE-LOCK:
    DELETE List.
  END.

  ASSIGN iIterOrder = 0
         iCurOrder = 0.

END. /* PROCEDURE Clear List */

/* -------------------------- End Methods List ------------------------*/


